__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2022, Vanessa Sochat"
__license__ = "MPL 2.0"

from citelang.main import Client


def main(args, parser, extra, subparser):

    cli = Client(quiet=args.quiet)
    result = cli.credit(
        name=args.package[1],
        manager=args.package[0],
        use_cache=not args.no_cache,
        max_depth=args.max_depth,
        max_deps=args.max_depth,
        min_credit=args.min_credit,
        credit_split=args.credit_split,
    )

    if args.json and not args.outfile:
        result.print_json()

    elif args.outfile:
        result.save(args.outfile)

    else:
        result.print_result()



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2022, Vanessa Sochat"
__license__ = "MPL 2.0"

__version__ = "0.0.28"
AUTHOR = "Vanessa Sochat"
EMAIL = "vsoch@users.noreply.github.com"
NAME = "citelang"
PACKAGE_URL = "https://github.com/vsoch/citelang"
KEYWORDS = "credit, software, rseng, research software"
DESCRIPTION = "Credit parser and markdown language for scientific software."
LICENSE = "LICENSE"

################################################################################
# Global requirements

# Since we assume wanting Singularity and lmod, we require spython and Jinja2

INSTALL_REQUIRES = (
    ("rich", {"min_version": None}),
    ("pyaml", {"min_version": None}),
    ("jsonschema", {"min_version": None}),
    ("requests", {"min_version": None}),
    ("colour", {"min_version": None}),
)

BADGE_REQUIRES = (
    ("plotly", {"min_version": None}),
    ("kaleido", {"min_version": None}),
)

TESTS_REQUIRES = (("pytest", {"min_version": "4.6.2"}),)

################################################################################
# Submodule Requirements (versions that include database)

INSTALL_REQUIRES_ALL = INSTALL_REQUIRES + TESTS_REQUIRES + BADGE_REQUIRES



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2022, Vanessa Sochat"
__license__ = "MPL 2.0"

from citelang.logger import logger
import citelang.utils as utils
import citelang.main.result as results
import citelang.main.settings as settings

import os
import shutil


class Cache:
    """
    The cache controls saving and loading citelang package content. We should
    be able to load (and use) a cache anywhere.
    """

    _cache = {}

    def __repr__(self):
        return str(self)

    def __str__(self):
        return "[citelang-cache]"

    def clear(self, force=False):
        """
        Clear the cache (with confirmation).
        """
        if not force and not utils.confirm_action(
            "Are you sure you want to clear the cache? "
        ):
            return
        if os.path.exists(settings.cfg.cache_dir):
            shutil.rmtree(settings.cfg.cache_dir)

    def set(self, name, result):
        """
        Given a result, cache if the user has cache enabled.
        """
        if settings.cfg.disable_cache == True:
            return

        # If we are using the memory cache, return from there.
        if not settings.cfg.disable_memory_cache:
            if name not in self._cache:
                self._cache[name] = result.data

            # If we end in a version, add to cache too
            # E.g., package/pypi/numpy/1.22.3
            if name.count("/") == 3:
                without_version = name.rsplit("/", 1)[0]
                self._cache[without_version] = result.data

        # Ensure cache directory exists
        utils.mkdir_p(settings.cfg.cache_dir)

        # prepare the path (e.g., cache_dir/package_managers.json)
        path = self.get_cache_name(name)

        # Don't write empty data
        if not result.data:
            logger.warning("No data found for result, not writing %s" % path)
            return

        # If we are using the memory cache, save to it
        if not settings.cfg.disable_memory_cache:
            self._cache[name] = result.data

        # We can't predict nesting, so always make directory
        utils.mkdir_p(os.path.dirname(path))
        utils.write_json(result.data, path)

    def get_cache_name(self, name):
        """
        Return a json cache entry.
        """
        return os.path.join(settings.cfg.cache_dir, "%s.json" % name)

    def is_empty(self, name):
        """
        Given a package name, determine if it's empty (the endpoint tried and
        no result) so we don't try again.
        """
        empty_name = os.path.join(settings.cfg.cache_dir, "%s.empty" % name)
        return os.path.exists(empty_name)

    def mark_empty(self, name):
        """
        Given a package name, mark it empty to indicate the manager doesn't
        have it.
        """
        empty_name = os.path.join(settings.cfg.cache_dir, "%s.empty" % name)
        utils.write_file("", empty_name)

    def get(self, name, endpoint=None):
        """
        Given a cache name (typically matching the endpoint) retrieve if exists.
        If provided and endpoint, wrap the result with the endpoint. Otherwise,
        return the json result.
        """
        # First effort - get from memory
        if name in self._cache:
            return self._cache[name]

        # Now look for filesystem
        path = self.get_cache_name(name)
        if not os.path.exists(path):
            return

        # Load the cache, return as a result if it exists.
        # If there is an error loading it, assume corrupt (and regnerate)
        data = None
        try:
            data = utils.read_json(path)
        except:
            logger.warning(f"Cache entry {path} has corrupt json, removing.")
            os.remove(path)

        if data and endpoint:
            return results.Table(data, endpoint)
        elif data:
            return data


def init_cache():
    return Cache()


cache = init_cache()



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2022, Vanessa Sochat"
__license__ = "MPL 2.0"

from citelang.logger import logger
import citelang.utils as utils
import citelang.main.graph as graph
import citelang.main.colors as colors

from rich.table import Table as RichTable
from rich.console import Console

import json
import os

here = os.path.dirname(os.path.abspath(__file__))


class Result:
    """
    A result holds the request result and can parse or return in different formats.
    """

    def __init__(self, data, endpoint=None):
        self.endpoint = endpoint

        # Does the endpoint want to sort or otherwise order?
        data = data or {}
        if endpoint:
            self.data = self.endpoint.order(data)
        else:
            self.data = data

    def print_json(self):
        print(utils.print_json(self.data))

    def to_dict(self):
        return self.data

    def save(self, outfile):
        """
        Save to output file
        """
        logger.info("Saving to %s..." % outfile)
        utils.write_json(self.data, outfile)


class Table(Result):
    """
    A table is a result formatted as a table for the client.
    """

    def __init__(self, data, endpoint=None):
        super().__init__(data, endpoint)

        # Keep track of the max length for each field not truncated
        self.max_widths = {}
        self.ensure_complete()

    def available_width(self, columns):
        """
        Calculate available width based on fields we cannot truncate (urls)
        """
        # We will determine column width based on terminal size
        try:
            width = os.get_terminal_size().columns
        except OSError:
            width = 120

        # Calculate column width
        column_width = int(width / len(columns))
        updated = width

        for _, needed in self.max_widths.items():
            updated = updated - needed

        # We don't have enough space
        if updated < 0:
            logger.warning("Terminal is too small to correctly render!")
            return column_width

        # Otherwise, recalculate column width taking into account truncation
        # We use the updated smaller width, and break it up between columns
        # that don't have a max width
        return int(updated / (len(columns) - len(self.max_widths)))

    def ensure_complete(self):
        """
        If any data missing fields, ensure they are included
        """
        if isinstance(self.data, list):
            self.ensure_complete_list()
        # We don't check other types for now

    def ensure_complete_list(self):
        """
        Given a list result, check the fields.
        """
        fields = set()
        for entry in self.data:
            [fields.add(x) for x in entry.keys()]

        # Ensure fields are present
        truncate_list = self.endpoint.truncate_list if self.endpoint else []
        for entry in self.data:
            for field in fields:

                if field not in entry:
                    entry[field] = ""

                # We can't truncate this field!
                if field in truncate_list:

                    # We haven't seen it before
                    if field not in self.max_widths:
                        self.max_widths[field] = len(entry[field])
                    elif self.max_widths[field] < len(entry[field]):
                        self.max_widths[field] = len(entry[field])

    def table_columns(self):
        """
        Shared function to return consistent table columns
        """
        # An endpoint that returns structured json can choose to flatten
        data = self.endpoint.table_data(self.data) if self.endpoint else self.data

        # Plan to remove empty columns with count 0
        column_counts = {x: 0 for x, _ in data[0].items()}

        # Count entries for each column
        for entry in data:
            for column, value in entry.items():
                if value:
                    column_counts[column] += 1

        # Get column titles
        columns = []
        skip_list = self.endpoint.skip_list if self.endpoint else []
        contenders = list(data[0].keys())
        for column in contenders:
            if column in skip_list or column_counts[column] == 0:
                continue
            columns.append(column)
        return columns

    def table_rows(self, columns, limit=25):
        """
        Shared function to yield rows as a list
        """
        # All keys are lowercase
        column_width = self.available_width(columns)

        # An endpoint that returns structured json can choose to flatten
        data = self.endpoint.table_data(self.data) if self.endpoint else self.data
        for i, row in enumerate(data):

            # have we gone over the limit?
            if limit and i > limit:
                return

            parsed = []
            truncate_list = self.endpoint.truncate_list if self.endpoint else []
            for column in columns:
                content = str(row[column]) if row[column] else ""
                if (
                    content
                    and len(content) > column_width
                    and column not in truncate_list
                ):
                    content = content[:column_width] + "..."
                parsed.append(content)
            yield parsed

    def table(self, limit=25):
        """
        Pretty print a table of results.
        """
        table_title = ""
        if isinstance(self.data, dict) and "name" in self.data:
            table_title = " " + self.data["name"]

        endpoint_name = (
            ""
            if not self.endpoint
            else self.endpoint.name.capitalize().replace("_", " ")
        )
        table = RichTable(title="%s%s" % (endpoint_name, table_title))

        # No dependencies!
        data = self.endpoint.table_data(self.data) if self.endpoint else self.data
        if not data:
            print("This package does not have any dependencies.")
            return

        # Get column titles and unique colors
        columns = self.table_columns()
        column_colors = colors.get_rich_colors(len(columns))

        for i, column in enumerate(columns):
            title = " ".join([x.capitalize() for x in column.split("_")])
            table.add_column(title, style=column_colors[i])

        # Add rows
        for row in self.table_rows(columns, limit=limit):
            table.add_row(*row)

        # And print!
        console = Console()
        console.print(table, justify="center")


class Tree(Result):
    def __init__(self, result):
        self.result = result
        self.parse_data()

    def parse_data(self):
        """
        Parse result into data. The size is for d3, and makes each node
        relative to its siblings.
        """
        # Generate colors for all nodes
        colorset = colors.get_color_range(N=self.result.total_nodes)
        color_lookup = {
            c: str(colorset[i]) for i, c in enumerate(self.result.children_names)
        }

        data = {
            "name": self.result.name,
            "credit": self.result.credit,
            "weight": self.result.weight,
            "size": 100,
            "round_by": self.result.round_by,
            "credit_rounded": round(self.result.credit, self.result.round_by),
            "children": [],
            "color": color_lookup[self.result.name],
        }

        def add(data, children, total, levels=1):
            total += data.credit
            if data.children:
                child_size = 100 / len(data.children)
                levels += 1
            for child in data.children:
                if child.name not in color_lookup:
                    color_lookup[child.name] = colors.get_random_color()
                node = {
                    "name": child.name,
                    "credit": child.credit,
                    "credit_rounded": round(child.credit, child.round_by),
                    "weight": child.weight,
                    "size": child_size,
                    "round_by": child.round_by,
                    "children": [],
                    "color": color_lookup[child.name],
                }
                total, levels = add(child, node["children"], total, levels)
                children.append(node)
            return total, levels

        # Get total and number of levels in the tree
        total, levels = add(self.result, data["children"], 0)
        self.data = data
        self.data["total"] = total
        self.data["levels"] = levels

    def print_result(self):
        """
        Print a tree result (todo redo with color and rich)
        """

        def print_result(result, indent=2):
            print(
                "%s%20s: %s"
                % (
                    " " * indent,
                    result["name"],
                    round(result["credit"], result["round_by"]),
                )
            )
            for child in result["children"]:
                print_result(child, indent=indent * 2)

        print_result(self.data)
        print("total: %s" % round(self.data["total"], 3))


class Graph(Result):
    """
    A graph result can generate text for a graph
    """

    def __init__(self, root):
        self.root = root

    def graph(self, fmt=None):
        """
        Generate a graph of dependencies
        """
        # Select output format (default to console)
        if fmt == "dot":
            out = graph.Dot(self.root)
        elif fmt == "cypher":
            out = graph.Cypher(self.root)
        elif fmt == "gexf":
            out = graph.Gexf(self.root)
        else:
            # Only console supported for now!
            out = graph.Console(self.root)
        return out.generate()


class Treemap(Tree):
    """
    A treemap is an interactive tree.
    """

    def print_result(self):
        pass

    def save(self, outfile):
        """
        Save to output file
        """
        logger.info("Saving to %s..." % outfile)
        template = utils.read_file(os.path.join(here, "badge", "treemap.html"))
        template = template.replace("{{title}}", self.result.name)

        # Add an extra parent to the data (root) so the one child is requests
        root = {"name": "citelang", "size": 100, "children": [self.data]}
        template = template.replace("{{data}}", json.dumps(root))
        utils.write_file(template, outfile)
        logger.info("Result saved to %s" % outfile)


class Badge(Tree):
    """
    This is an static badge that uses plotly.
    """

    def print_result(self):
        pass

    def save(self, outfile):
        """
        Save to output file
        """
        import citelang.main.badge as badge

        logger.info("Saving to %s..." % outfile)

        # Force root credit value to be sum of children in first layer
        self.data["credit_label"] = self.data["credit"]
        total_credit = 0
        for child in self.data.get("children", []):
            total_credit += child["credit"]
        self.data["credit"] = total_credit

        # Add an extra parent to the data (root) so the one child is requests
        badge.generate(
            self.data,
            outfile,
            min_credit=self.result.min_credit,
            credit_split=self.result.credit_split,
        )
        logger.info("Result saved to %s" % outfile)


class InteractiveBadge(Tree):
    """
    A badge uses tree data with d3 for an interactive visualization
    """

    def print_result(self):
        pass

    def save(self, outfile):
        """
        Save to output file
        """
        logger.info("Saving to %s..." % outfile)
        template = utils.read_file(os.path.join(here, "badge", "sunburst.html"))
        template = template.replace("{{title}}", self.result.name)

        # Remove root node weight
        del self.data["weight"]

        # Add an extra parent to the data (root) so the one child is requests
        template = template.replace("{{data}}", json.dumps(self.data))
        utils.write_file(template, outfile)
        logger.info("Result saved to %s" % outfile)



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2022, Vanessa Sochat"
__license__ = "MPL 2.0"

from citelang.logger import logger

import random
import sys

from .base import GraphBase


class Dot(GraphBase):
    """
    The dot format is for graphviz
    """

    def generate(self, graphname=None, fontname="Arial"):
        if self.outfile != sys.stdout:
            logger.info("Output will be written to %s" % self.outfile)
            fd = open(self.outfile, "w")
        else:
            fd = self.outfile

        graphname = graphname or self.data.name
        fd.write("digraph " + graphname + " {\n ratio=0.562;\n")

        # Do we want to render using a certain font?
        for node in ["graph", "node", "edge"]:
            fd.write(" " + node + ' [fontname="' + fontname + '"];\n')

        # Color for root (first) and then linked libs, and symbols
        colors = [
            "#" + "".join([random.choice("0123456789ABCDEF") for j in range(6)])
            for i in range(2)
        ]

        root_color = colors.pop()
        node_color = colors.pop()

        # Create graph with nodes
        next_nodes = [self.data]
        first_node = True
        while next_nodes:
            next_node = next_nodes.pop(0)
            label = self.uids[next_node.name]
            name = "%s (%s)" % (next_node.name, round(next_node.credit, 3))
            if first_node:
                fd.write(
                    ' %s [label="%s" tooltip="%s", style=filled, color="%s"];\n'
                    % (label, name, name, root_color)
                )
                first_node = False
            else:
                fd.write(
                    ' %s [label="%s" tooltip="%s", style=filled, color="%s"];\n'
                    % (label, name, name, node_color)
                )
            for child in next_node.children:
                next_nodes.append(child)

        # Now create links
        next_nodes = [self.data]
        while next_nodes:
            next_node = next_nodes.pop(0)
            label = self.uids[next_node.name]
            name = "%s (%s)" % (next_node.name, round(next_node.credit, 3))
            for child in next_node.children:
                label_to = self.uids[child.name]
                fd.write(
                    ' %s -> %s [label=" depends on " tooltip="%s -> %s"];\n'
                    % (label, label_to, next_node.name, child.name)
                )
                next_nodes.append(child)

        fd.write("\n}\n")
        if self.outfile != sys.stdout:
            fd.close()



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2022, Vanessa Sochat"
__license__ = "MPL 2.0"

# Custom package managers not in libraries IO

import requests
import jsonschema
from .base import PackageManager
import citelang.main.schemas as schemas
from citelang.logger import logger


class SpackManager(PackageManager):
    """
    Packages from spack.
    """

    name = "spack"
    apiroot = "https://spack.github.io/packages/data"
    homepage = "https://spack.github.io/packages"
    color = "#0f3a80"
    default_language = None
    default_versions = []

    @property
    def project_count(self):
        try:  # "pythonic"
            return len(requests.get("%s/packages.json" % self.apiroot).json())
        except:
            return None

    def dependencies(self, name):
        """
        Get dependencies for a spack package

        Some package managers have separate endpoints for this, but we use
        the same package endpoint and return that data.
        """
        if not hasattr(self, "data"):
            self.package(name)
        return self.data

    def package(self, name, **kwargs):
        """
        Get metadata for a spack package. Try to format like libraries.io
        """
        url = "%s/packages/%s.json" % (self.apiroot, name)
        response = requests.get(url)
        if response.status_code != 200:
            logger.exit("There was an issue retrieving %s" % url)

        # The spack meta response already has name, description, homepage, versions
        meta = response.json()
        meta["keywords"] = []
        meta["language"] = None
        meta["versions"] = [{"number": x["name"]} for x in meta["versions"]]

        # Ensure we match schema
        jsonschema.validate(meta, schema=schemas.package)
        self.data = meta
        return meta



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2022, Vanessa Sochat"
__license__ = "MPL 2.0"


class Node:
    def __init__(
        self, obj, weight, credit_split=0.5, depth=0, min_credit=0.01, is_root=False
    ):
        """
        A graph node has a name (or object), weight, and children
        """
        self.obj = obj
        self.weight = weight
        self.children = []
        self.credit_split = credit_split
        self.depth = depth
        self.min_credit = min_credit
        self.is_root = is_root

        # If we stop parsing we can set this credit to whatever the node and
        # children would originall get
        self.total_credit = None
        self.children_names = []

    def iternodes(self):
        """
        Iterate through all nodes of the graph
        """
        nodes = [self]
        while nodes:
            node = nodes.pop(0)
            yield node
            nodes += node.children

    @property
    def total_nodes(self):
        return len(self.children_names)

    @property
    def name(self):
        return self.obj.name

    @property
    def round_by(self):
        value = str(self.min_credit).split(".", 1)[-1]
        if not value:
            return 3
        return len(value)

    def add_child(self, child):
        """
        Add child ensures we increment the depth by 1
        """
        child.depth = self.depth + 1
        self.children.append(child)

    @property
    def credit(self):
        if self.total_credit:
            return self.total_credit
        return self.credit_split * self.weight


class Graph:
    def __init__(self, root):
        self.root = root



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2022, Vanessa Sochat"
__license__ = "MPL 2.0"

import os
import citelang.utils as utils

apiroot = "https://libraries.io"

install_dir = utils.get_installdir()

# User home
user_home = os.path.expanduser("~/")

# CiteLang default home
citelang_home = os.path.join(user_home, ".citelang")

# Replacements allowed
reps = {
    "$install_dir": install_dir,
    "$root_dir": os.path.dirname(install_dir),
    "$user_home": user_home,
    "$citelang_home": citelang_home,
}

# The default settings file in the install root
default_settings_file = os.path.join(reps["$install_dir"], "settings.yml")

# Optionally can come from environment
env_settings_file = os.environ.get("CITELANG_SETTINGS_FILE")

# Don't try using pip
use_pip = False if os.environ.get("CITELANG_USE_PIP") == "false" else True

# The user settings file can be created to over-ride default
user_settings_file = os.path.join(citelang_home, "settings.yml")

# variables in settings that allow environment variable expansion
allowed_envars = ["cache_dir"]



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2022, Vanessa Sochat"
__license__ = "MPL 2.0"


from citelang.logger import logger
import citelang.defaults as defaults
import citelang.main.schemas as schemas
import citelang.utils
import shutil

import jsonschema
import os
import re
import yaml


class SettingsBase:
    def __init__(self, settings_file=None, validate=True):
        """
        Create a new settings object, which requires a settings file to load
        """
        self.settings_file = None
        self.user_settings = None
        self.load(settings_file)
        if validate:
            self.validate()

    def __str__(self):
        return "[citelang-settings]"

    def __repr__(self):
        return self.__str__()

    def validate(self):
        """
        Validate the loaded settings with jsonschema
        """
        jsonschema.validate(instance=self._settings, schema=schemas.settings)

    def inituser(self):
        """
        Create a user specific config in user's home.
        """
        user_home = os.path.dirname(defaults.user_settings_file)
        if not os.path.exists(user_home):
            os.makedirs(user_home)
        if os.path.exists(defaults.user_settings_file):
            logger.exit(
                "%s already exists! Remove first before re-creating."
                % defaults.user_settings_file
            )
        shutil.copyfile(self.settings_file, defaults.user_settings_file)
        logger.info("Created user settings file %s" % defaults.user_settings_file)

    def edit(self):
        """
        Interactively edit a config file.
        """
        if not self.settings_file or not os.path.exists(self.settings_file):
            logger.exit("Settings file not found.")

        # Make sure editor exists first!
        editor = citelang.utils.which(self.config_editor)
        if editor["return_code"] != 0:
            logger.exit(
                "Editor '%s' not found! Update with citelang config set config_editor:<name>"
                % self.config_editor
            )
        citelang.utils.run_command(
            [self.config_editor, self.settings_file], stream=True
        )

    def get_settings_file(self, settings_file=None):
        """
        Get the preferred user settings file, set user settings if exists.
        """
        # Always use environment first
        env_settings = None
        if defaults.env_settings_file and os.path.exists(defaults.env_settings_file):
            env_settings = defaults.env_settings_file

        # Only consider user settings if the file exists!
        user_settings = None
        if os.path.exists(defaults.user_settings_file):
            user_settings = defaults.user_settings_file

        # First preference to env_settings, then user settings, then default
        return env_settings or user_settings or defaults.default_settings_file

    def load(self, settings_file=None):
        """
        Load the settings file into the settings object
        """
        # Get the preferred settings flie
        self.settings_file = self.get_settings_file(settings_file)

        # Exit quickly if the settings file does not exist
        if not os.path.exists(self.settings_file):
            logger.exit("%s does not exist." % self.settings_file)

        # Always load default settings first
        with open(defaults.default_settings_file, "r") as fd:
            self._settings = yaml.load(fd.read(), Loader=yaml.SafeLoader)

        # Update with user or custom settings if not equal to default
        if self.settings_file != defaults.default_settings_file:
            with open(self.settings_file, "r") as fd:
                self._settings.update(yaml.load(fd.read(), Loader=yaml.SafeLoader))

    def get(self, key, default=None):
        """
        Get a settings value, doing appropriate substitution and expansion.
        """
        # This is a reference to a dictionary (object) setting
        if ":" in key:
            key, subkey = key.split(":")
            value = self._settings[key][subkey]
        else:
            value = self._settings.get(key, default)
        value = self._substitutions(value)
        # If we allow environment substitution, do it
        if key in defaults.allowed_envars and value:
            if isinstance(value, list):
                value = [os.path.expandvars(v) for v in value]
            else:
                value = os.path.expandvars(value)
        return value

    def __getattr__(self, key):
        """
        A direct get of an attribute, but default to None if doesn't exist
        """
        return self.get(key)

    def add(self, key, value):
        """
        Add a value to a list parameter
        """
        value = self.parse_boolean(value)

        # We can only add to lists
        current = self._settings.get(key)
        if current and not isinstance(current, list):
            logger.exit("You cannot only add to a list variable.")
        value = self.parse_null(value)

        if value not in current:
            # Add to the beginning of the list
            current = [value] + current
            self._settings[key] = []
            [self._settings[key].append(x) for x in current]
            self.change_validate(key, value)
            logger.warning(
                "Warning: Check with citelang config edit - ordering of list can change."
            )

    def remove(self, key, value):
        """
        Remove a value from a list parameter
        """
        current = self._settings.get(key)
        if current and not isinstance(current, list):
            logger.exit("You cannot only remove from a list variable.")
        if not current or value not in current:
            logger.exit("%s is not in %s" % (value, key))
        current.pop(current.index(value))
        self._settings[key] = current
        self.change_validate(key, current)
        logger.warning(
            "Warning: Check with citelang config edit - ordering of list can change."
        )

    def parse_boolean(self, value):
        """
        If the value is True/False, ensure we return a boolean
        """
        if isinstance(value, str) and value.lower() == "true":
            value = True
        elif isinstance(value, str) and value.lower() == "false":
            value = False
        return value

    def parse_null(self, value):
        """
        Given a null or none from the command line, ensure parsed as None type
        """
        if isinstance(value, str) and value.lower() in ["none", "null"]:
            return None

        # Ensure we strip strings
        if isinstance(value, str):
            value = value.strip()
        return value

    def set(self, key, value):
        """
        Set a setting based on key and value. If the key has :, it's nested
        """
        value = self.parse_boolean(value)

        # List values not allowed for set
        current = self._settings.get(key)
        if current and isinstance(current, list):
            logger.exit("You cannot use 'set' for a list. Use add/remove instead.")

        # This is a reference to a dictionary (object) setting
        if isinstance(value, str) and ":" in value:
            subkey, value = value.split(":")
            value = self.parse_boolean(value)
            value = self.parse_null(value)
            self._settings[key][subkey] = value
        else:
            value = self.parse_null(value)
            self._settings[key] = value

        # Validate and catch error message cleanly
        self.change_validate(key, value)

    def change_validate(self, key, value):
        """
        A courtesy function to validate a new config addition.
        """
        # Don't allow the user to add a setting not known
        try:
            self.validate()
        except jsonschema.exceptions.ValidationError as error:
            logger.exit(
                "%s:%s cannot be added to config: %s" % (key, value, error.message)
            )

    def _substitutions(self, value):
        """
        Given a value, make substitutions
        """
        if isinstance(value, bool) or not value:
            return value

        # Currently dicts only support boolean or null so we return as is
        elif isinstance(value, dict):
            return value

        for rep, repvalue in defaults.reps.items():
            if isinstance(value, list):
                value = [x.replace(rep, repvalue) for x in value]
            else:
                value = value.replace(rep, repvalue)

        return value

    def delete(self, key):
        if key in self._settings:
            del self._settings[key]

    def save(self, filename=None):
        """
        Save settings, but do not change order of anything.
        """
        filename = filename or self.settings_file
        if not filename:
            logger.exit("A filename is required to save to.")

        with open(filename, "w") as fd:
            yaml.dump(self._settings, fd)

    def __iter__(self):
        for key, value in self.__dict__.items():
            yield key, value

    def update_params(self, params):
        """
        Update a configuration on the fly (no save) only for set/add/remove.
        Unlike the traditional set/get/add functions, this function expects
        each entry in the params list to start with the action, e.g.:

        set:name:value
        add:name:value
        rm:name:value
        """
        # Cut out early if params not provided
        if not params:
            return

        for param in params:
            if not re.search("^(add|set|rm)", param, re.IGNORECASE) or ":" not in param:
                logger.warning(
                    "Parameter update request must start with (add|set|rm):, skipping %s"
                )
            command, param = param.split(":", 1)
            self.update_param(command.lower(), param)

    def update_param(self, command, param):
        """
        Given a parameter, update the configuration on the fly if it's in set/add/remove
        """
        if ":" not in param:
            logger.warning(
                "Param %s is missing a :, should be key:value pair. Skipping." % param
            )
            return

        key, value = param.split(":", 1)
        if command == "set":
            self.set(key, value)
            logger.info("Updated %s to be %s" % (key, value))
        elif command == "add":
            self.add(key, value)
            logger.info("Added %s to %s" % (key, value))
        elif command == "remove":
            self.remove(key, value)
            logger.info("Removed %s from %s" % (key, value))


class Settings(SettingsBase):
    """
    The settings class is a wrapper for easily parsing a settings.yml file.

    We parse into a query-able class. It also gives us control to update settings,
    meaning we change the values and then write them to file. It's basically
    a dictionary-like class with extra functions.
    """

    def __init__(self, settings_file, validate=True):
        """
        Create a new settings object, which requires a settings file to load
        """
        self.load(settings_file)
        if validate:
            self.validate()


# Global settings
def init_settings(validate=True):
    return SettingsBase(validate=validate)


cfg = init_settings()



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2022, Vanessa Sochat"
__license__ = "MPL 2.0"

from .base import PackagesFromFile
import re


class GemfileManager(PackagesFromFile):
    """
    Packages from a ruby Gemfile
    """

    name = "Gemfile"
    underlying_manager = "rubygems"
    default_language = "ruby"
    project_count = None
    homepage = "https://rubygems.org/"
    color = "#e9573f"
    default_versions = None

    def parse(self, content, **kwargs):
        """
        Parse the self.content (the Gemfile)
        """
        repo = self.get_repo()
        libs = []

        for line in content.split("\n"):
            line = line.strip()
            if line.startswith("gem") and "source" not in line:
                for sub in ["gem", "'", '"', ","]:
                    line = line.replace(sub, "").strip()

                for sep in [" if ", ":platforms", ":require"]:
                    line = line.split(sep)[0]

                # Do we have a version?
                if re.search("(~>|@|<=|>=)", line):
                    line, _, version = re.split("(~>|@|<=|>=)", line)
                    line = "%s@%s" % (line.strip(), version.strip())
                libs.append(line)

        deps = []
        for package_name in libs:
            version = None
            if "@" in package_name:
                package_name, version = package_name.split("@", 1)
            elif " " in package_name:
                package_name, version = package_name.split(" ", 1)

            pkg = self.get_package(package_name, version)
            if not pkg:
                continue

            # Ensure we have version, fallback to latest
            if not version:
                version = pkg.data["latest_release_number"]

            # Require saving to cache here - many expensive calls
            cache_name = f"package/rubygems/{package_name}/{version}"
            self.cache.set(cache_name, pkg)

            # use latest release version. This will be wrong for an old
            # dependency, but it's not worth it to make a ton of extra API calls
            dep = {
                "name": package_name,
                "project_name": package_name,
                "number": version,
                "published_at": pkg.data["latest_stable_release_published_at"],
                "researched_at": None,
                "spdx_expression": "NOASSERTION",
                "original_license": pkg.data["licenses"],
                "repository_sources": ["Rubygems"],
            }
            deps.append(dep)

        repo["dependencies"] = deps
        self.data["package"] = repo
        self.data["dependencies"] = deps
        return repo



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2022, Vanessa Sochat"
__license__ = "MPL 2.0"

from citelang.logger import logger
import citelang.main.cache as cache
import citelang.main.endpoints as endpoints
import citelang.main.result as results
import citelang.main.graph as graph
import citelang.main.package as package
import citelang.main.packages as packages


class BaseClient:
    """
    A baseclient controls interactions with endpoints and the cache.
    """

    def __init__(self, quiet=False, **kwargs):
        self.quiet = quiet
        self.cache = cache.cache

    def __repr__(self):
        return str(self)

    def __str__(self):
        return "[citelang-client]"

    def dependencies(self, manager, name, use_cache=True):
        """
        Get dependencies for a package. If no version, use latest.
        """
        self.check_manager(manager, use_cache)
        pkg = package.Package(manager, name, use_cache=use_cache)
        return pkg.dependencies()

    def package_managers(self, use_cache=True):
        """
        Get a listing of package managers supported on libraries.io
        """
        result = self.cache.get("package_managers")
        if not result or not use_cache:
            logger.info("Retrieving new result for package managers...")
            result = endpoints.get_endpoint("package_managers")

            # Update custom package managers
            names = [x["name"] for x in result.data]
            result.data += [
                p().info()
                for _, p in packages.managers.items()
                if p().info()["name"] not in names
            ]
        else:
            result = endpoints.get_endpoint("package_managers", data=result)

        # If cache is enabled, we save the result
        self.cache.set("package_managers", result)
        return result

    def check_manager(self, name, use_cache=True):
        """
        Check if a package manager is supported
        """
        if name in packages.manager_names:
            return
        managers = self.package_managers(use_cache)

        # Ensure the manager is allowed
        if managers.data:
            managers = [x["name"].lower() for x in managers.data]
            if name not in managers and name is not None:
                managers = "\n".join(managers)
                logger.exit(f"{name} is not a known manager. Choices are:\n{managers}")

    def graph(self, fmt=None, *args, **kwargs):
        """
        Generate a graph for a package.

        credit_split is how to split credit between some package and its dependents. E.g., 0.5 means 50/50. 0.8 means
        the main package gets 80%, and dependencies split 20%. We go up until the min credit 0.05 at which case we
        stop adding.
        """
        root = self._graph(*args, **kwargs)
        return results.Graph(root).graph(fmt=fmt)

    def badge(self, template="static", *args, **kwargs):
        """
        Generate a badge for a package
        """
        root = self._graph(*args, **kwargs)
        if template == "treemap":
            return results.Treemap(root)
        elif template == "sunburst":
            return results.InteractiveBadge(root)
        return results.Badge(root)

    def _graph(
        self,
        manager,
        name,
        use_cache=True,
        max_depth=None,
        max_deps=None,
        min_credit=0.01,
        credit_split=0.01,
        pkg=None,
    ):
        """
        Shared 'private' function to generate graph
        """
        self.check_manager(manager, use_cache)

        # Allow the caller to provide a pre-generated (often custom) package
        if not pkg:
            pkg = package.get_package(manager, name, use_cache=use_cache)

        # E.g., a requirements.txt file underlying manager is pypi
        if pkg.underlying_manager:
            manager = pkg.underlying_manager.underlying_manager

        # keep track of deps (we only care about name, not version) and names
        seen = set()
        node_names = set()

        # Top node gets full credit 1.0 (to split between itself and deps)
        root = graph.Node(
            obj=pkg,
            weight=1.0,
            credit_split=credit_split,
            depth=0,
            min_credit=min_credit,
            is_root=True,
        )

        # A pointer to the next node
        next_nodes = [root]

        # Booleans to trigger exiting parser
        stop_looking = False

        # Keep handle to previous children in case we stop looking
        previous = []

        while next_nodes and not stop_looking:
            next_node = next_nodes.pop(0)

            # Sometimes we know a package name / version but cannot get deps
            try:
                deps = next_node.obj.dependencies(return_data=True)
            except:
                deps = []
            [
                node_names.add(d["name"])
                for d in deps
                if "name" in d and not d["name"].startswith("__")
            ]

            # Stopping point - exceeded max depth
            if max_depth and next_node.depth > max_depth:
                stop_looking = True

            # Stopping point - exceeded max deps (plus 1 for original package)
            if max_deps and len(seen) + 1 > max_deps:
                stop_looking = True

            seen.add(next_node.name)
            node_names.add(next_node.name)

            if not deps:
                continue

            # How much credit for each of the new deps?
            # credit for all deps AS a percentage of the weight BROKEN into number of deps
            dep_credit = ((1 - credit_split) * next_node.weight) / len(deps)

            # Stopping point - dependency credit is too small
            if dep_credit < min_credit:
                stop_looking = True

            # If we are stopping, time to break and distribute remaining credit
            # to nodes we couldn't parse children for.
            if stop_looking:

                # We haven't parsed this one yet
                next_node.total_credit = next_node.weight

                # Give all remaining credit to the node we aren't parsing
                for node in previous:

                    # We might have added and given credit to some children
                    if node.children:

                        # Redistribute credit amongst children that were > threshold
                        dep_credit = ((1 - credit_split) * node.weight) / len(
                            node.children
                        )
                        for child in node.children:
                            child.total_credit = dep_credit

                        # The node's credit is that weight minus total dep credit
                        node.total_credit = node.weight - (
                            dep_credit * len(node.children)
                        )
                    else:
                        node.total_credit = node.weight
                break

            # Calculate credit for each dependency and add as child
            for dep in deps:

                # Haven't seen this case, but just a check
                dep_name = dep["name"] or dep["project_name"]
                if not dep_name or dep_name.startswith("__"):
                    continue

                depnode = package.get_package(manager, dep_name, use_cache=use_cache)
                child = graph.Node(
                    obj=depnode,
                    weight=dep_credit,
                    credit_split=credit_split,
                    depth=next_node.depth + 1,
                    min_credit=min_credit,
                )
                next_node.add_child(child)
                next_nodes.append(child)

                # Store previous if we need to update weight
                previous.append(child)

        # Add total node count to root
        root.children_names = list(node_names)
        return root

    def credit(self, *args, **kwargs):
        """
        Get the credit root node, then do additional graph parsing.
        """
        root = self._graph(*args, **kwargs)
        return results.Tree(root)

    def package(self, manager, name, use_cache=True):
        """
        Lookup a package in a specific package manager
        """
        self.check_manager(manager, use_cache)
        pkg = package.get_package(manager, name, use_cache=use_cache)
        return pkg.info()



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2022, Vanessa Sochat"
__license__ = "MPL 2.0"

import citelang.main.client as client
from citelang.logger import logger
import tempfile
import os


def main(args, parser, extra, subparser):
    if os.path.isfile(args.package[1]):
        cli = client.get_parser(filename=args.package[1], quiet=args.quiet)
        result = cli.badge(
            name=args.package[0],
            use_cache=not args.no_cache,
            max_depth=args.max_depth,
            max_deps=args.max_depth,
            min_credit=args.min_credit,
            credit_split=args.credit_split,
            template=args.template,
        )

    # Case 2: named package and manager
    else:
        cli = client.Client(quiet=args.quiet)
        result = cli.badge(
            name=args.package[1],
            manager=args.package[0],
            use_cache=not args.no_cache,
            max_depth=args.max_depth,
            max_deps=args.max_depth,
            min_credit=args.min_credit,
            credit_split=args.credit_split,
            template=args.template,
        )

    if not args.outfile and args.template == "static":
        args.outfile = os.path.join(
            "%s-%s.png" % (args.package[0], args.package[1])
        ).replace(os.sep, "-")

    elif not args.outfile:
        args.outfile = os.path.join(tempfile.mkdtemp(), "index.html")

    if os.path.exists(args.outfile) and not args.force:
        logger.exit(f"{args.outfile} exists, add --force to overwrite.")
    result.save(args.outfile)



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2022, Vanessa Sochat"
__license__ = "MPL 2.0"

import citelang.main.client as client
import citelang.utils as utils
from citelang.logger import logger
import citelang.main.result as results


def main(args, parser, extra, subparser):
    parser = client.ContributionParser(
        root=args.root, start=args.start, end=args.end, outdir=args.outdir
    )

    # All time true means we include git blames done before the range
    summary = parser.parse(within_range=not args.all_time)

    if args.by == "author":
        data = summary.by_author(args.detail, filters=args.filters)
    else:
        data = summary.by_file(args.detail, filters=args.filters)

    # Do we want to save data to file?
    if args.outfile:
        logger.info("Saving to %s" % args.outfile)
        utils.write_json(data, args.outfile)
        return

    # Otherwise, parse into table
    if args.detail:

        # Separate paths by newlines
        for item in data:
            item["paths"] = "\n".join(
                ["%s: %s" % (k, v) for k, v in item["paths"].items()]
            )
        table = results.Table(data)
    else:
        table = results.Table(data)
    table.table(limit=None)



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2022, Vanessa Sochat"
__license__ = "MPL 2.0"

from citelang.logger import logger
from .base import PackagesFromFile


class CMakeListManager(PackagesFromFile):
    """
    Packages from a CMakeLists.txt
    """

    name = "CMakeLists.txt"

    # We use spack because we can probably find a lot of C++ deps here!
    underlying_manager = "spack"
    default_language = "C++"
    project_count = None
    homepage = None
    color = "#e9573f"
    default_versions = None

    def parse(self, content, **kwargs):
        """
        Attempt to parse the CMakeList.txt file. This can give us at least
        one level of dependencies.
        """
        repo = self.get_repo()
        libs = []

        for line in content.split("\n"):
            line = line.strip()

            # A version could be here as second item...
            if "find_package" in line:
                lib = line.split("(")[1].split(" ")[0].strip().replace(")", "")
                if not lib:
                    continue
                logger.info(f"Found lib {lib}")
                libs.append(lib.lower())

            elif "cmake_minimum_required" in line:
                version = line.split("(")[1].split(" ")[1].strip(")").strip()
                libs.append(f"cmake@{version}")

            elif "set(PROJECT_VERSION" in line:
                version = (
                    line.split(" ")[1]
                    .split(" ")[0]
                    .replace(")", "")
                    .replace('"', "")
                    .replace("'", "")
                )
                repo["versions"][0]["number"] = version

            elif "project" in line:
                repo["name"] = line.split("(")[1].split(" ")[0].strip()

        deps = []
        for package_name in libs:
            version = None
            if "@" in package_name:
                package_name, version = package_name.split("@", 1)

            # use latest release version. This will be wrong for an old
            # dependency, but it's not worth it to make a ton of extra API calls
            dep = {
                "name": package_name,
                "project_name": package_name,
                "number": version,
                "published_at": repo["versions"][0]["published_at"],
                "researched_at": None,
                "spdx_expression": "NOASSERTION",
                "original_license": None,
                "repository_sources": ["Cpp"],
            }
            deps.append(dep)

        repo["dependencies"] = deps
        self.data["package"] = repo
        self.data["dependencies"] = deps
        return repo



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2022, Vanessa Sochat"
__license__ = "MPL 2.0"


from citelang.logger import logger
import os
import json
import requests
import time

default_headers = {"Accept": "application/json", "User-Agent": "citelang-python"}
params = {"per_page": 100}

api_key = os.environ.get("CITELANG_LIBRARIES_KEY")
if api_key:
    params.update({"api_key": api_key})


def check_response(typ, r, return_json=True, stream=False, retry=True):
    """
    Ensure the response status code is 20x
    """
    session = requests.session()

    # Rate is 60/minute
    if r.status_code == 429:
        logger.info("Exceeded API limit, sleeping 1 minute.")
        time.sleep(60)
        r = session.send(r.request)
        return check_response(typ, r, return_json, stream, retry=retry)

    if r.status_code == 401:
        logger.exit("You must set CITELAG_LIBRARIES_KEY in the environment.")

    if r.status_code not in [200, 201]:
        logger.exit(
            "Unsuccessful response: %s, %s %s" % (r.status_code, r.reason, r.text)
        )

    # All data is typically json
    if return_json and not stream:
        return r.json()
    return r


def print_response(r):
    """
    Print the result of a response
    """
    response = r.json()
    logger.info("%s: %s" % (r.url, json.dumps(response, indent=4)))


def do_request(
    typ,
    url,
    data=None,
    json=None,
    headers=None,
    return_json=True,
    stream=False,
):
    """
    Do a request (get, post, etc)
    """
    # If we have a cached token, use it!
    headers = headers or {}
    headers.update(default_headers)
    logger.info("%s %s" % (typ.upper(), url))

    # The first post when you upload the model defines the flavor (regression)
    if json:
        r = requests.request(
            typ, url, json=json, headers=headers, stream=stream, params=params
        )
    else:
        r = requests.request(
            typ, url, data=data, headers=headers, stream=stream, params=params
        )
    if not stream and not return_json:
        print_response(r)
    return check_response(typ, r, return_json=return_json, stream=stream)


def post(url, data=None, json=None, headers=None, return_json=True):
    """
    Perform a POST request
    """
    return do_request(
        "post", url, data=data, json=json, headers=headers, return_json=return_json
    )


def delete(url, data=None, json=None, headers=None, return_json=True):
    """
    Perform a DELETE request
    """
    return do_request(
        "delete",
        url,
        data=data,
        json=json,
        headers=headers,
        return_json=return_json,
    )


def get(url, data=None, json=None, headers=None, return_json=True, stream=False):
    """
    Perform a GET request
    """
    return do_request(
        "get",
        url,
        data=data,
        json=json,
        headers=headers,
        return_json=return_json,
        stream=stream,
    )



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2022, Vanessa Sochat"
__license__ = "MPL 2.0"

# Custom package managers not in libraries IO

import sys
from citelang.logger import logger
import citelang.main.cache as cache
import citelang.main.endpoints as endpoints
import citelang.utils as utils
import requests


class PackageManager:
    """
    A package manager is a custom class (not supported by libraries io)
    """

    underlying_manager = None
    filesystem_manager = False

    def __init__(self, *args, **kwargs):
        self.cache = cache.cache
        self.data = {}
        for attr in [
            "name",
            "project_count",
            "homepage",
            "color",
            "default_language",
            "default_versions",
        ]:
            if not hasattr(self, attr):
                sys.exit(
                    "Misconfigured package manager %s missing %s attribute"
                    % (self.name, attr)
                )

    # If needed, make sure endpoint data is sorted
    def info(self):
        return {
            "name": self.underlying_manager or self.name,
            "project_count": self.project_count,
            "homepage": self.homepage,
            "color": self.color,
            "default_language": self.default_language,
        }

    def get_or_fail(self, url, headers=None, return_text=False):
        """
        Shared endpoint to get a url or fail.
        """
        response = requests.get(url, headers=headers)
        if response.status_code != 200:
            logger.exit("Cannot retrieve %s: %s" % (url, response.json()))
        if return_text:
            return response.text
        return response.json()

    def package(self, name, **kwargs):
        raise NotImplementedError


class PackagesFromFile(PackageManager):
    """
    Load packages from file. The class requires a parse function to parse
    the content provided.
    """

    filesystem_manager = True

    def __init__(self, package_name=None, content=None, filename=None):
        """
        An base filesystem package manager parses packages from a file
        """
        self.version = None
        self.package_name = package_name
        super().__init__()
        if package_name:
            self.set_name(package_name)
        self.data = {}
        self.filename = filename
        if content:
            self.parse(content)

    def set_name(self, name):
        if "@" in name:
            name, version = name.split("@", 1)
            self.version = version
        if "[" in name and "]" in name:
            name = name.split("[", 1)[0]

        # invalid characters found!
        name = name.replace(";", "")
        self.package_name = name

    def get_package(self, package_name, version=None):
        """
        Shared function to get a package based on name, version
        """
        # Try to get from cache - either versioned or not
        pkg = None
        if package_name and version:
            cache_name = f"package/{self.underlying_manager}/{package_name}/{version}"
            result = self.cache.get(cache_name)
            if result:
                pkg = endpoints.get_endpoint("package", data=result)

        elif package_name and not version:
            cache_name = f"package/{self.underlying_manager}/{package_name}"
            result = self.cache.get(cache_name)
            if result:
                pkg = endpoints.get_endpoint("package", data=result)

        if pkg is None:

            # Don't try again if this package is flagged as empty (not existing)
            if self.cache.is_empty(f"package/{self.underlying_manager}/{package_name}"):
                return pkg
            try:
                pkg = endpoints.get_endpoint(
                    "package",
                    package_name=package_name,
                    manager=self.underlying_manager,
                )
            except:
                pass

        # If we get down here and no package, mark empty (unless it isn't a package we know)
        if not pkg:
            try:
                self.cache.mark_empty(
                    f"package/{self.underlying_manager}/{package_name}"
                )
            except:
                pass
        return pkg

    def get_repo(self):
        """
        Helper function to return start of repo metadata. Intended to be used
        in the self.parse() function.
        """
        # The "repo" is the package name, we can't be sure about versions
        versions = [
            {
                "published_at": utils.get_time_now(),
                "number": self.version or "latest",
            }
        ]
        repo = {"name": self.package_name, "versions": versions}

        # Try to provide a default version
        repo["default_version"] = self.version or "latest"
        return repo

    def package(self, name, **kwargs):
        """
        The package endpoint ignores the name and just returns parsed data
        """
        return self.data.get("package")

    def parse(self, content, **kwargs):
        """
        Parse the self.content
        """
        raise NotImplementedError



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2022, Vanessa Sochat"
__license__ = "MPL 2.0"

from citelang.logger import logger
import citelang.utils as utils
import citelang.main.base as base
import citelang.main.result as results
import citelang.main.package as package
import citelang.main.packages as packages
from operator import itemgetter

import re
import os

table_template = """|Manager|Name|Credit|
|-------|----|------|
"""

template_suffix = "\n> Note that credit values are rounded and expanded (so shared dependencies are represented as one record) and may not add to 1.0. Rounded values that hit zero are removed.\n"

citation_regex = "\@([a-zA-Z0-9]+)\{(.*?)\}"


class Parser(base.BaseClient):
    """
    A parser reads in a markdown file, finds software references, and generates
    a software graph per the user preferences.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.data = {}
        self.libs = []
        self.round_by = 3
        self.roots = {}

    def gen(self, name, manager, *args, **kwargs):
        """
        Generate a one off credit table for a named library
        """
        # Prepare a parser that can generate a table
        self.add_lib(name=name, manager=manager)
        self._update_roots(**kwargs)
        self.prepare_table()
        return self

    def parse(self, **kwargs):
        """
        A parser is specific to a base class.
        """
        raise NotImplementedError

    def _update_roots(self, **kwargs):
        """
        Update roots for new libraries we've found
        """
        for lib in self.libs:
            if "name" not in lib or "manager" not in lib:
                logger.warning("Skipping %s, missing name or manager." % lib)
            uid = "%s:%s" % (lib["manager"], lib["name"])

            # Don't generate roots we've already parsed
            if uid in self.roots:
                continue

            if "version" in lib or "release" in lib:
                version = lib.get("version") or lib.get("release")
                uid = "%s@%s" % (uid, version)
                lib["name"] = "%s@%s" % (lib["name"], version)
            self.roots[uid] = self._graph(
                manager=lib["manager"], name=lib["name"], **kwargs
            )

    def prepare_table(self, *args, **kwargs):
        """
        Given a package name and manager, generate roots and a table
        """
        # Generate the table with multiple roots - flatten out credit
        table, round_by = self._prepare_table(self.roots)

        # Add listing of packages and dependencies to parser
        self.data = table
        self.round_by = round_by
        return self

    def _prepare_table(self, roots):
        """
        Helper function to prepare the average credit summary table.
        """
        table = {}

        # Multiplier for credit depending on total packages
        splitby = 1 / len(roots)
        for lib, root in roots.items():
            for node in root.iternodes():
                if node.obj.manager not in table:
                    table[node.obj.manager] = {}
                if node.name not in table[node.obj.manager]:
                    table[node.obj.manager][node.name] = {
                        "credit": 0,
                        "url": node.obj.homepage,
                    }
                table[node.obj.manager][node.name]["credit"] += node.credit * splitby
        return table, root.round_by

    def prepare_custom_table(
        self,
        includes,
    ):
        """
        Prepare (and return) a table filtered to a specific kind of root.
        """
        # Multiplier for credit depending on total packages
        roots = {}
        for lib, root in self.roots.items():
            reqfile = lib.split(":")[0]
            if reqfile in includes:
                roots[lib] = root

        table, _ = self._prepare_table(roots)
        return table

    def add_lib(self, manager, name, **args):
        """
        Manually add a library (e.g., not reading from a pre-existing file)
        """
        self.libs.append({"manager": manager, "name": name, **args})


class FileNameParser(Parser):
    """
    FileNameParser is an abstract parser for one that requires an input filename.
    """

    def __init__(self, filename: str = None, *args, **kwargs):
        self.filename = None
        self.content = None

        # Are we rendering back into the raw content?
        self.rendering_content = True

        super().__init__(*args, **kwargs)
        if filename:
            self.load(filename)
            self.filename = os.path.abspath(filename)

    def load(self, filename):
        if not os.path.exists(filename):
            logger.exit(f"{filename} does not exist")
        self.content = utils.read_file(filename)

    @property
    def start_block(self):
        return "<!--citelang start-->"

    @property
    def end_block(self):
        return "<!--citelang end-->"

    @property
    def empty_content(self):
        return """# Software Credit

<!--citelang start-->
<!--citelang end-->

- Generated by [CiteLang](https://github.com/vsoch/citelang)
"""

    def check(self):
        """
        Checks for citelang.
        """
        if self.start_block not in self.content:
            logger.exit(
                "Cannot find %s, ensure it is present where you want the citelang table."
                % self.start_block
            )
        if self.end_block not in self.content:
            logger.exit(
                "Cannot find %s, ensure it is present where you want the citelang table."
                % self.end_block
            )

    def parse(self, **kwargs):
        """
        Given a markdown file, return a list of parse packages and versions.
        """
        self.check()
        for match in re.findall(citation_regex, self.content):
            if len(match) != 2:
                logger.warning(
                    "found malformed citation reference %s, skipping." % match
                )
                continue
            args = {
                y[0]: y[1]
                for y in [x.strip().split("=") for x in match[-1].split(",")]
                if len(y) == 2
            }
            self.libs.append({"manager": match[0], **args})

        # Generate roots from the libraries we have
        self._update_roots(**kwargs)
        return self.prepare_table()

    def render(self, start_end_blocks=True, data=None):
        """
        Render final file!
        """
        markdown = table_template

        # Allow parsing custom data
        data = data or self.data

        # Sort from least to greatest
        listing = []
        for manager, pkgs in data.items():
            for pkg, meta in pkgs.items():
                listing.append((manager, pkg, meta["credit"], meta["url"]))

        listing = sorted(listing, key=itemgetter(2), reverse=True)
        for (manager, pkg, credit, url) in listing:
            credit = round(credit, self.round_by)
            if credit == 0:
                logger.warning("Rounded credit for %s is 0, skipping." % pkg)
                continue

            # R packages have some "doubled up" urls, by space or newline
            if url:
                if re.search("( |\n)", url):
                    url = re.split("( |\n)", url.strip())[0]
                pkg = "[%s](%s)" % (pkg, url)
            markdown += "|%s|%s|%s|\n" % (
                manager,
                pkg,
                credit,
            )

        if not self.rendering_content:
            content = self.empty_content
        else:
            content = self.content or self.empty_content

        render = []
        lines = content.split("\n")
        while lines:
            line = lines.pop(0)
            if self.start_block in line:
                while self.end_block not in line:
                    line = lines.pop(0)
                # When we get here we have the end block

                if start_end_blocks:
                    render += (
                        [self.start_block]
                        + markdown.split("\n")
                        + [template_suffix, self.end_block]
                    )
                else:
                    render += markdown.split("\n") + [template_suffix]

            else:
                render.append(line)
        return "\n".join(render)


class RequirementsParser(FileNameParser):
    """
    Given one or more requirements files, create packages for summary.
    """

    def __init__(self, filename: str = None, *args, **kwargs):
        super().__init__(filename=filename, *args, **kwargs)

        # We don't want to render back into requirements.txt
        self.rendering_content = False

    def badge(self, name, template="static", filename=None, *args, **kwargs):
        """
        Generate a badge for a requirements file.
        """
        self.gen(filename=filename or self.filename, name=name, *args, **kwargs)
        root = list(self.roots.values())[0]
        if template == "treemap":
            return results.Treemap(root)
        elif template == "sunburst":
            return results.InteractiveBadge(root)
        return results.Badge(root)

    def load_datafiles(self, files, includes=None):
        """
        Given a list of data.json files, load into roots and then generate the table.
        """
        roots = {}

        # Take average depending on number of roots
        splitby = 1 / len(files)
        for filename in files:
            if not os.path.exists(filename):
                logger.warning("%s does not exist, skipping." % filename)
                continue
            data = utils.read_json(filename)
            for manager, libs in data.items():
                if includes and manager not in includes:
                    continue
                if manager not in roots:
                    roots[manager] = {}
                for libname, libmeta in libs.items():
                    if not libname:
                        continue
                    if libname not in roots[manager]:
                        libmeta["credit"] = libmeta["credit"] * splitby
                        roots[manager][libname] = libmeta
                    else:
                        if libmeta["url"] and not roots[manager][libname]["url"]:
                            roots[manager][libname]["url"] = libmeta["url"]
                        roots[manager][libname]["credit"] += libmeta["credit"] * splitby
        return roots

    def gen(self, name, filename=None, *args, **kwargs):
        """
        Add a requirements file. Manager is determined by filename, and name
        is required to label the package.
        """
        if filename is not None:
            self.load(filename)
        else:
            filename = self.filename

        if not self.content:
            logger.exit("no filename provided or filename is empty, nothing to parse.")

        # Do we have a known dependency file?
        basename = os.path.basename(filename)
        pkg = None

        if basename in packages.filesystem_manager_names:
            manager_kwargs = {
                "content": self.content,
                "package_name": name,
                "filename": filename,
            }

            # Custom set the name of the manager
            pkg = package.get_package(
                manager=basename,
                name=name,
                manager_kwargs=manager_kwargs,
            )
            # Populate dependencies and package
            pkg.info()

            uid = "%s:%s" % (basename, filename)
            self.roots[uid] = self._graph(
                manager=pkg.underlying_manager.underlying_manager,
                name=name,
                pkg=pkg,
                **kwargs,
            )

        if not pkg:
            logger.exit(f"Dependency file type {basename} not known, or none found.")

        self.prepare_table()
        return self



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2022, Vanessa Sochat"
__license__ = "MPL 2.0"

from .base import PackagesFromFile


class GoModuleManager(PackagesFromFile):
    """
    Packages from a go module file
    """

    name = "go.mod"
    underlying_manager = "go"
    default_language = "Go"
    project_count = None
    homepage = "https://go.dev/"
    color = "#007d9c"
    default_versions = None

    def parse(self, content, **kwargs):
        """
        Parse the self.content (the Gemfile)
        """
        repo = self.get_repo()

        libs = []
        parsing = False
        for line in content.split("\n"):
            if "require (" in line:
                parsing = True
                continue
            if parsing and ")" in line:
                break
            if parsing:
                line = line.strip()

                # Get rid of comments
                line = line.split("//")[0].strip()
                libs.append(line)

        deps = []
        for package_name in libs:
            version = None
            if " " in package_name:
                package_name, version = package_name.split(" ", 1)

            # Many go packages are not known
            pkg = None
            try:
                pkg = self.get_package(package_name, version)
            except:
                pass

            if pkg is None:
                dep = {
                    "name": package_name,
                    "project_name": package_name,
                    "number": version,
                    "published_at": None,
                    "researched_at": None,
                    "spdx_expression": "NOASSERTION",
                    "original_license": None,
                    "repository_sources": ["Unknown"],
                }
                deps.append(dep)
                continue

            # Ensure we have version, fallback to latest
            if not version:
                version = dep.data["latest_release_number"]

            # Require saving to cache here - many expensive calls
            cache_name = f"package/go/{package_name}/{version}"
            self.cache.set(cache_name, pkg)

            # use latest release version. This will be wrong for an old
            # dependency, but it's not worth it to make a ton of extra API calls
            dep = {
                "name": package_name,
                "project_name": package_name,
                "number": version,
                "published_at": dep.data["latest_stable_release_published_at"],
                "researched_at": None,
                "spdx_expression": "NOASSERTION",
                "original_license": pkg.data["licenses"],
                "repository_sources": ["Go"],
            }
            deps.append(dep)

        repo["dependencies"] = deps
        self.data["package"] = repo
        self.data["dependencies"] = deps
        return repo



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2022, Vanessa Sochat"
__license__ = "MPL 2.0"

from citelang.logger import logger
import citelang.main.colors as colors


try:
    import plotly.graph_objs as go
except ImportError:
    logger.exit(
        "Plotly and dependencies are needed to generate this badge. pip install citelang[badge]."
    )


def generate(data, outfile=None, min_credit=0.01, credit_split=0.5, height=800):
    """
    Generate a static badge
    """
    labels, parents, values = unwrap_tree(data)
    ids, parentids = get_ids(labels, parents)

    # colors for subtrees
    colorset = colors.get_color_range(N=max(data["levels"] - 1, 1))

    # add color for root (white)
    colorset = ["#ffffff"] + [str(x) for x in colorset]

    trace = go.Sunburst(
        ids=ids,
        labels=labels,
        parents=parentids,
        values=values,
        marker={"colors": colorset, "line": {"width": 1.5}},
        branchvalues="total",
    )

    layout = go.Layout(
        title={
            "text": "Credit allocation for %s<br><sup>The total credit graph adds to ~1, and cutoff is %s</sup><br><sup>All children split %s of the parent credit</sup>"
            % (data["name"], min_credit, credit_split),
            "x": 0.5,
        },
        font={"size": 12},
        showlegend=False,
        autosize=False,
        height=height,
        xaxis={"visible": False},
        yaxis={"visible": False},
        hovermode="closest",
    )

    fig = go.Figure(data=[trace], layout=layout)
    if outfile:
        fig.write_image(outfile)
    return fig


def get_ids(labels, parents):
    """
    Given labels and parent node ids, generate labels for the nodes.
    """
    if len(labels) != len(parents):
        logger.exit(
            "The list of labels should have the same length like the list of parents"
        )
    ids = [str(id) for id in range(len(labels))]

    # associate to each label the corresponding id
    dlabels = {label: idx for label, idx in zip(labels, ids)}

    # empty is for the root node
    parentids = [""] + [dlabels[label] for label in parents[1:]]
    return ids, parentids


def unwrap_tree(tree_dict, labels=None, parents=None, values=None):
    """
    Given a typical structured data with nodes and children, unwrap into flat lists.
    """
    labels = labels or []
    parents = parents or [""]
    values = values or []
    credit = tree_dict.get("credit_label") or tree_dict.get("credit")
    node_name = "%s<br>%s" % (
        tree_dict["name"],
        round(credit, tree_dict["round_by"]),
    )
    labels.append(node_name)
    values.append(tree_dict["credit"])
    for child in tree_dict.get("children"):
        parents.append(node_name)
        unwrap_tree(child, labels, parents, values)
    return labels, parents, values



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2022, Vanessa Sochat"
__license__ = "MPL 2.0"

from citelang.logger import logger
import citelang.defaults as defaults
import citelang.main.http as http
import citelang.main.result as results
from datetime import datetime
import sys

# Registered endpoints (populated on init)
registry = {}
registry_names = []


def get_endpoint(name, data=None, **kwargs):
    """
    Get a named endpoint, optionally, using the cache (default)
    """
    if name not in registry:
        names = registry_names
        logger.exit(f"{name} is not a known endpoint. Choose from {names}")

    # If we get a manual manager, change to it's apppropriate package
    if "manager" in kwargs and kwargs["manager"] == "requirements.txt":
        kwargs["manager"] = "pypi"

    # Create the endpoint with any optional params
    if not data:
        endpoint = registry[name](**kwargs, require_params=False)
        return results.Table(http.get(endpoint.url), endpoint)
    endpoint = registry[name](**kwargs)
    return results.Table(data, endpoint)


class Endpoint:
    dont_truncate = ["url"]
    emoji = "sparkles"

    def __init__(self, require_params=False, *args, **kwargs):
        self.apiroot = defaults.apiroot
        for attr in ["name", "path"]:
            if not hasattr(self, attr):
                sys.exit(
                    "Misconfigured endpoint %s missing %s attribute" % (self, attr)
                )

        # If we have attribute format_url we require the params
        if require_params:
            self.require_params(**kwargs)
        else:
            self.format_params(**kwargs)

    def table_data(self, data):
        """
        A custom data parser, optionally if an endpoint needs to parse data further.
        """
        return data

    def require_params(self, **kwargs):
        required = getattr(self, "format_url", [])
        params = {}
        for require in required:
            if require not in kwargs:
                sys.exit(f"parameter {require} is required for this endpoint.")
            params[require] = kwargs[require]

        # If we have parameters, format the path
        if params:
            self.format_params(**params)
        self.params = params

    def format_params(self, **params):
        # If we have parameters, format the path
        if params:
            self.path = self.path.format(**params)
        self.params = params

    # If needed, make sure endpoint data is sorted
    def order(self, data):
        return data

    @property
    def url(self):
        return self.apiroot + self.path

    @property
    def title(self):
        return self.name.replace("_", " ").capitalize()

    def _get_attribute_or_list(self, name):
        listing = []
        if hasattr(self, name):
            listing = getattr(self, name, []) or []
        return listing

    @property
    def truncate_list(self):
        return self._get_attribute_or_list("dont_truncate")

    @property
    def skip_list(self):
        return self._get_attribute_or_list("skips")


class PackageManagers(Endpoint):
    name = "package_managers"
    path = "/api/platforms"
    emoji = "box"
    skips = ["color"]

    def table_data(self, data):
        """
        Ensure we don't include any repeats and sort by name
        """
        seen = set()
        final_set = []
        for item in data:
            if item["name"].lower() in seen:
                continue

            # Ensure custom managers are capitalized for display
            item["name"] = item["name"].capitalize()
            final_set.append(item)
            seen.add(item["name"].lower())
        return sorted(final_set, key=lambda x: x["name"])


class Dependencies(Endpoint):
    name = "dependencies"
    path = "/api/{manager}/{package_name}/{version}/dependencies"
    format_url = ["manager", "package_name", "version"]
    emoji = "arrow"
    skips = ["normalized_licenses"]
    dont_truncate = ["project_name"]

    def table_data(self, data):
        return data.get("dependencies")


class Package(Endpoint):
    name = "package"
    path = "/api/{manager}/{package_name}"
    emoji = "box"
    format_url = ["manager", "package_name"]
    skips = [
        "repository_sources",
        "spdx_expression",
        "researched_at",
        "original_license",
    ]

    def order(self, data):
        """
        Order versions by published at
        """
        versions = data.get("versions", [])
        if versions and "published_at" in versions[0]:
            try:
                data["versions"] = sorted(
                    versions,
                    key=lambda x: datetime.strptime(
                        x["published_at"].split("T", 1)[0], "%Y-%m-%d"
                    ),
                )
            except:
                pass

        if "versions" not in data:
            data["versions"] = sorted(
                data["versions"], key=lambda x: x["number"], reverse=True
            )

        return data

    def table_data(self, data):
        # package should return a list of versions
        return data.get("versions")

    @property
    def title(self):
        return "Package " + self.params.get("package_name", "")


for endpoint in [PackageManagers, Package, Dependencies]:
    registry_names.append(endpoint.name)
    registry[endpoint.name] = endpoint

registry_names.sort()



#!/usr/bin/python

import os

from citelang.main.settings import Settings

here = os.path.dirname(os.path.abspath(__file__))
root = os.path.dirname(here)


def test_environment_substitution(tmp_path):
    """
    Test variable substitution
    """
    settings_file = os.path.join(root, "settings.yml")
    settings = Settings(settings_file)

    assert "pancakes" not in settings.cache_dir
    assert "cache" in settings.cache_dir
    os.environ["SOME_PATH"] = "/tmp"
    os.putenv("SOME_PATH", "/tmp")
    settings.set("cache_dir", "$SOME_PATH/pancakes")
    assert settings.cache_dir == "/tmp/pancakes"


def test_set_get(tmp_path):
    """
    Test variable set/get
    """
    settings_file = os.path.join(root, "settings.yml")
    settings = Settings(settings_file)
    assert settings.cache_dir.endswith("cache")
    settings.set("cache_dir", "/tmp/cache")
    assert settings.cache_dir == "/tmp/cache"



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2022, Vanessa Sochat"
__license__ = "MPL 2.0"

from pip._internal.commands.install import (
    InstallCommand,
    reject_location_related_install_options,
)
from pip._internal.cli.cmdoptions import make_target_python
from pip._internal.cli.main_parser import create_main_parser
from pip._internal.req.req_tracker import get_requirement_tracker
from pip._internal.utils.temp_dir import TempDirectory
from pip._internal.cli import cmdoptions
from pip._internal.cache import WheelCache

from citelang.logger import logger
import citelang.utils as utils
import re
import os

from .base import PackagesFromFile


class PackageLister(InstallCommand):
    """
    A wrapper to the pypi install command that simply resolves dependencies
    and saves the listing. A requirements.txt file will work outside of context,
    however a setup.py typically needs external dependencies (and if it does
    not work, will fall back to the static parser).
    """

    def run(self, options, args):
        cmdoptions.check_dist_restriction(options, check_target=True)
        options.use_user_site = True
        session = self.get_default_session(options)
        target_python = make_target_python(options)
        finder = self._build_package_finder(
            options=options,
            session=session,
            target_python=target_python,
            ignore_requires_python=options.ignore_requires_python,
        )
        try:
            reqs = self.get_requirements(args, options, finder, session)
            self._citelang_success = True

            # requirements.txt will be parsed already
            if "." not in args:
                self._citelang_reqs = reqs
                return 0

            # If we are installing a setup.py, we need to parse
            for req in reqs:
                req.permit_editable_wheels = True

            req_tracker = self.enter_context(get_requirement_tracker())
            wheel_cache = WheelCache(options.cache_dir, options.format_control)
            reject_location_related_install_options(reqs, options.install_options)
            directory = TempDirectory(
                delete=True,
                kind="install",
                globally_managed=True,
            )
            preparer = self.make_requirement_preparer(
                temp_build_dir=directory,
                options=options,
                req_tracker=req_tracker,
                session=session,
                finder=finder,
                use_user_site=options.use_user_site,
                verbosity=self.verbosity,
            )
            resolver = self.make_resolver(
                preparer=preparer,
                finder=finder,
                options=options,
                wheel_cache=wheel_cache,
                use_user_site=options.use_user_site,
                ignore_installed=options.ignore_installed,
                ignore_requires_python=options.ignore_requires_python,
                force_reinstall=options.force_reinstall,
                upgrade_strategy="to-satisfy-only",
                use_pep517=options.use_pep517,
            )
            self.trace_basic_info(finder)
            self._citelang_reqs = resolver.resolve(
                reqs, check_supported_wheels=False
            ).all_requirements

        except Exception as err:
            self._citelang_success = False
            logger.warning(
                "Issue using pip to resolve dependencies: %s, will fall back to static parser."
                % err
            )
            self._citelang_reqs = []

        # Always return success to main
        return 0


class PipManager(PackagesFromFile):
    def parse_python_deps(self, lines):
        """
        Shared function for deriving a list of python dependencies from lines
        """
        parser = create_main_parser()
        args = []
        if self.filename.endswith("requirements.txt"):
            # Generate args based on a faux command
            args = ["install", "-r", "requirements.txt"]
        if self.filename.endswith("setup.py"):
            args = ["install", "."]

        general_options, args_else = parser.parse_args(args)
        cmd_name = args_else[0]
        cmd_args = args[:]
        cmd_args.remove(cmd_name)

        # Instead of parsing the content we instead parse the filename
        dirname = os.path.dirname(self.filename)
        lister = PackageLister("citelang", "listing of Citelang packages")

        with utils.workdir(dirname):
            lister.main(cmd_args)

        # If we don't have deps, fall back to static parser
        if not lister._citelang_success:
            from .python_base import PythonManager as FallbackParser

            FallbackParser.name = self.name
            mgr = FallbackParser(package_name=self.package_name, filename=self.filename)
            return mgr.parse_python_deps(lines=lines, filename=self.filename)

        deps = []
        for dep in lister._citelang_reqs:
            package_name = dep.name
            version = None

            # We can only use package names that could be on pip
            if not package_name:
                continue
            if dep.specifier:
                version = str(dep.specifier)
            if version and re.search("(==|~=|<=|>=|<|>|!=)", version):
                version = re.sub("(==|~=|<=|>=|<|>|!=)", "", version).strip()

            # First add requirements (names and pypi manager) to deps
            pkg = self.get_package(package_name, version)
            if not pkg:
                pkg = self.get_package(package_name)
                if not pkg:
                    logger.warning("Issue getting package %s, skipping" % package_name)
                    continue

            # Ensure we have version, fallback to latest
            if not version:
                version = pkg.data["latest_release_number"]

            # Require saving to cache here - many expensive calls
            cache_name = f"package/pypi/{package_name}/{version}"
            self.cache.set(cache_name, pkg)

            # use latest release version. This will be wrong for an old
            # dependency, but it's not worth it to make a ton of extra API calls
            dep = {
                "name": package_name,
                "project_name": package_name,
                "number": version,
                "published_at": pkg.data["latest_stable_release_published_at"],
                "researched_at": None,
                "spdx_expression": "NOASSERTION",
                "original_license": pkg.data["licenses"],
                "repository_sources": ["Pypi"],
            }
            deps.append(dep)
        return deps
